#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"
#include "texture.h"

// These two external function pointers are exported by the VU code
// and we can now use them as pointers to the compiled VU data
extern "C" void VU_vu1code_start(void);
extern "C" void VU_vu1code_end(void);

#define ROTYMAX (PIHALF - 0.1f)

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

// Position the camera
static Vector4	CamPos(0.0f, 60.0f, 130.0f, 0);

Matrix4x4 GetCameraMatrix(void)
{
	static const float ROTXMAX  = (PIHALF - 0.1f);
		
	// Set up a nice 3D camera
	static float yrot = 0.0f;
	static float xrot = -0.57f;
	
	//CamPos.DumpVector4();
	//printf("%f, %f\n", xrot, yrot);
	
	// Reset the camer if requested
	if(pad[0].buttons & PAD_R2)
	{
		CamPos = Vector4(0.0f, 60.0f, 130.0f, 0);
		yrot = 0.0f;
		xrot = -0.57f;
	}
	
	// Get any requested rotations
	yrot -= pad[0].axes[PAD_AXIS_RX] * Abs(pad[0].axes[PAD_AXIS_RX]) * 0.07f;
	xrot += pad[0].axes[PAD_AXIS_RY] * Abs(pad[0].axes[PAD_AXIS_RY]) * 0.07f;
	
	// Clamp xrot to prevent going upside down.
	if(xrot > ROTXMAX)
		xrot = ROTXMAX;
	
	if(xrot < -ROTXMAX)
		xrot = -ROTXMAX;
	
	// Sort out the rotations
	Matrix4x4 roty, rotx;
	roty.RotationY( yrot );
	rotx.RotationX( xrot );
	Matrix4x4 CamRot = rotx * roty;
	
	// Transpose rotation matrix to get camera matrix
	CamRot = Transpose(CamRot);

	// Get the direction vectors for the camera
	Vector4 vXAxis = Vector4(CamRot(0,0), CamRot(1,0), CamRot(2,0), 1);
	Vector4 vYAxis = Vector4(CamRot(0,1), CamRot(1,1), CamRot(2,1), 1);
	Vector4 vZAxis = Vector4(CamRot(0,2), CamRot(1,2), CamRot(2,2), 1);
	
	// Get any requested translations
	float strafe =   0.5f * pad[0].axes[PAD_AXIS_LX];
	float forwards = 0.5f * pad[0].axes[PAD_AXIS_LY];
	float updown =   0.5f * (pad[0].pressures[PAD_PL1] - pad[0].pressures[PAD_PL2]);

	// Work out the requested move for this frame
	Vector4 vReqMove;
	vReqMove =  vXAxis * strafe;
	vReqMove += vYAxis * updown;
	vReqMove += vZAxis * forwards;
	
	// And add it to the camera position
	CamPos += vReqMove * 2.0f;

	// Finally build the camera matrix
	Matrix4x4 matView;
	matView.Translation(-CamPos.x, -CamPos.y, -CamPos.z);
	matView = matView * CamRot;
	
	return matView;
}


int main()
{
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total

	VIFStaticDMA.Initialise(3072);	// 2048 * 4K Pages = 8MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA

	// Two texture buffers, texbuf 0 = 480, texbuf 1 = 496
	int iCurrentTexBuf = 0;

	// Find out the length of the code in double words
	// We use double words since code is made up of many
	// pairs of two (1 upper, and 1 lower) VU instructions. 
	// Each instruction is 32 bits, so VU code must be a
	// multiple of 64 bits.
	uint32 iCodeLengthInDWORDs = (((uint32)&VU_vu1code_end - (uint32)&VU_vu1code_start) + 7) >> 3;
	// Copy the code to VU1_MICRO_MEM
	memcpy(VU1_MICRO_MEM, &VU_vu1code_start, iCodeLengthInDWORDs * 8);

	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);

	// Set up the DMA packet to clear the screen. We want to clear to blue.
	SPS2Manager.InitScreenClear(0, 0, 0x80);

	// Build a projection matrix
	float n = 2.0f;		// Near plane
	float f = 2000.0f;	// Far plane
	float FovYdiv2 = DegToRad(60.0f) * 0.5f; // 60 degree FOV
	float fAspect = 4.0f / 3.0f;			 // Aspect ratio
	float cotFOV = 1.0f / (Sin(FovYdiv2) / Cos(FovYdiv2));
	// We will be projecting to the 4096 wide drawing area, but we only want
	// the projection matrix to cover the visible area
	float w = cotFOV * (sps2UScreenGetWidth() / 4096.0f) / fAspect;
	float h = cotFOV * (sps2UScreenGetHeight() / 4096.0f);

	Matrix4x4 matProj
		(w, 0, 0, 0,
		 0, -h, 0, 0,
		 0, 0, (f+n) / (f-n), -1,
		 0, 0, (2*f*n) / (f-n), 0);

	Matrix4x4 matView;

	// Now add all of the data that won't change per-frame
	// to the static DMA buffer. The less work the CPU has to do
	// per-frame the better

	// Get the address that we can call to later.
	int iDMAInit = VIFStaticDMA.GetPointer();

	VIFStaticDMA.Add32(FLUSH);			// Make sure VU1 isn't busy
	VIFStaticDMA.Add32(STCYCL(1,1));	// Unpack linearly, i.e. don't skip any spaces
	VIFStaticDMA.Add32(BASE(16));		// The double buffers start at VU1 address 16 (giving us 16 QW to store data that won't change)
	VIFStaticDMA.Add32(OFFSET(504));	// The size of each buffer.

	VIFStaticDMA.AddUnpack(V4_32, 0, 2);	// Unpack 3QW of data that won't change

	VIFStaticDMA.AddVector(Vector4(2048, 2048, (0xFFFFFF / 32.0f), 1));	// 0: The scales
	VIFStaticDMA.AddVector(Vector4(1.0f, 1.0f, 1.0f, 1.0f));	// 1: The light colour

	VIFStaticDMA.DMARet();

	int iDMAPatch = VIFStaticDMA.GetPointer();

	// The verts are 3QW each, and the 504QW is split in half. Therefore we have 252QW for each buffer.
	// Each buffer also has 2QW of info (GIFTag + Numverts), so we can process 83 vertices (249QW)
	// in each buffer.
	int iNumVertsPerChunk = 83;

	// The number of triangles in each dimension.
	int iSizeX = 128;
	int iSizeY = 128;

	int iNumVerts = (iSizeX * 2) * (iSizeY - 1);

	int iX = 0, iY = 0;
	int bBottomRow = 0;
	bool bFirst = true;

	while(iNumVerts > 0)
	{
		int iVertsThisChunk;

		if(iNumVerts > iNumVertsPerChunk)
		{
			iVertsThisChunk = iNumVertsPerChunk;
			iNumVerts -= iVertsThisChunk;
		}
		else
		{
			iVertsThisChunk = iNumVerts;
			iNumVerts -= iVertsThisChunk;
		}

		VIFStaticDMA.AddUnpack(V4_32, 0, iVertsThisChunk * 3 + 2, 1);
		VIFStaticDMA.Add128(iVertsThisChunk * 3);
		VIFStaticDMA.Add128(
			GS_GIFTAG_BATCH(iVertsThisChunk,
					1, 0, 0, GIF_FLG_PACKED,
					GS_BATCH_3(GIF_REG_ST, GIF_REG_RGBAQ, GIF_REG_XYZ2)));

		// Zigzag from left to right across the patch making a huge triangle strip.
		//
		//	0	   	2		4		6
		//  |     / |     / |     / |
  		//  |   /   |   /   |   /   |
		//	1 /		3 /		5 /		7 etc.
		//
		for(int iVert = 0; iVert < iVertsThisChunk; iVert++)
		{
			float fX = (float)iX;
			float fY = (float)(iY + bBottomRow);

			// Disable the drawing kick when we move to the next row, so that we don't get a huge
			// polygon going across the patch.
			int iDontDraw = (iX == 0) ? 1 : 0;

			// Alternate between the current row and the one below.
			iX += bBottomRow;
			bBottomRow ^= 1;

			// If we have reached the end, start again on the next row.
			if(iX == iSizeX)
			{
				iX = 0;
				bBottomRow = 0;
				iY += 1;
			}

			// Add a huge number of vertices 
			VIFStaticDMA.AddVector(Vector4(fX / iSizeX, fY / iSizeY, 1, *((float *)&iDontDraw)));	// Tex coords + in W, a flag to set ADC or not.
			VIFStaticDMA.AddVector(Vector4(0, 1, 0, 0));											// Normal
			VIFStaticDMA.AddVector(Vector4((fX - (iSizeX / 2)), 0, (fY - (iSizeY / 2)), 1));		// Position
		}

		if(bFirst)
		{
			bFirst = false;
			// If this is the first buffer then run the VU code from address 0 so that the initialisation
			// VU code is ran.
			VIFStaticDMA.Add32(MSCALL(0));
		}
		else
		{
			// For multiple times MSCNT means to continue where the VU code finished last time at the --cont.
			VIFStaticDMA.Add32(MSCNT);
		}

		// Notice that due to the quad buffer setup we don't need to flush here because we can upload, process, and render
		// all in parallel now. (the XGKick in the vector code will flush to make sure that we don't get ahead of ourselves)
	}

	VIFStaticDMA.Add32(FLUSH);
	VIFStaticDMA.DMARet();

	float t = 0;
	Vector4 LightPos(0, 10, 0, 1);

	while(g_bLoop)
	{
		// Fire off the previous frame's buffer (or init data if it's the first time)
		VIFDynamicDMA.Fire();

		pad_update(PAD_0);
		if(pad[0].buttons & PAD_START && pad[0].buttons & PAD_SELECT)
			g_bLoop = false;


		// Build the camera matrix
		matView = GetCameraMatrix();
		
		// Clear the screen
		SPS2Manager.BeginScene();

		Matrix4x4 matWorld, matTrans;

		matWorld.Translation(0, 0, 0);
		matTrans = matWorld * matView * matProj;

		// Call the packet with the static data that doesn't change per frame.
		VIFDynamicDMA.DMACall(iDMAInit);

		// Update the point light position.
		t += 0.01f;
		LightPos.x = Sin( t ) * 60.0f;
		LightPos.z = Cos( t ) * 60.0f;

		// Add the dynamic point light position.
		VIFDynamicDMA.AddUnpack(V4_32, 2, 2);
		VIFDynamicDMA.AddVector(LightPos);
		
		// Add the camera position for specular highlights
		VIFDynamicDMA.AddVector(CamPos);

		// Now add the stuff that does change per frame
		VIFDynamicDMA.AddUnpack(V4_32, 8, 8);
		VIFDynamicDMA.AddMatrix(matTrans);

		// Add the world matrix for tranforming normals and light positions.
		VIFDynamicDMA.AddMatrix(matWorld);

		// Set the PRIM field here. This may look like it's for convenience, but actually it's the only
		// way we can do it. If we set this in the GIFTag it would reset the strip every buffer.
		// As we want one huge strip, we can't use the prim setting capabilites of the GIFTag.
		VIFDynamicDMA.StartDirect();
		VIFDynamicDMA.StartAD();
		VIFDynamicDMA.AddAD(PRIM_SET(GS_TRISTRIP, 1, 0, 0, 0, 0, 0, 0, 0), PRIM);
		VIFDynamicDMA.EndAD();
		VIFDynamicDMA.EndDirect();

		// Finally, render the patch!
		VIFDynamicDMA.DMACall(iDMAPatch);

		SPS2Manager.EndScene();
	}

	return 0;
}

